#!/usr/bin/env python
# encoding: utf-8
"""
@version: Python 3.7
@author: BingZhen Zhou
@contact: 953129171@qq.com
@file name: DG1Deasy.py
@time: 2019/7/30  10:27
"""

"""
DG method solve eq.
$u_x = f,\quad x\in [0,1];\quad u(0) = a.$
"""

import numpy as np
from numpy.polynomial import legendre
from numpy.linalg import solve

# k is the degree, k > 0
# Interplation points using the lobatto points
k = 3
nElements = np.array([10,20,40,80,160,320])
nElement = nElements[0]


# Get integral  points and weights
class Quadrature(object):
    def __init__(self, n):
        ideg = int(n)
        if ideg != n or n < 1:
            raise ValueError("n must be a non-negative integer")
        self.permuteData = n
        self.points = []
        self.weights = []

    # A standard  code to get the lobatto points and weights
    def getLobattoPiontsWeights(self):
        if self.permuteData == 2:
            self.points = np.array([-1, 1])
            self.weights = np.array([1, 1])
            return self.points, self.weights
        u = [np.sqrt(1.0*j/(2*j+1)*(j+2)/(2*j+3)) for j in range(1, self.permuteData-2)]
        bp = np.linalg.eigvals(np.diag(u, 1) + np.diag(u, -1))
        self.points = np.hstack(([-1], np.sort(bp), [1]))
        n = self.permuteData - 1
        lx = legendre.legval(self.points, [0]*n+[1])
        self.weights = 2.0/(n*lx*(n+1)*lx)
        return self.points, self.weights

    # A standard  code to get the gauss points and the weights
    def getGuassPointsWeights(self):
        self.points, self.weights = legendre.leggauss(self.permuteData)
        return self.points, self.weights

class Interp(object):
    def __init__(self, ip, quad):
        """
        ip: Interplation points
        quad: class of Quadrature
        """
        self.deg = len(ip)-1
        self.pp = self.deg + 1
        self.ip = ip
        self.quad = quad
        self.matrixH = self.mathbbH()
        self.matrixH_1 = self.mathbbH_1()

    def mathbbH(self):
        mat = np.zeros((self.pp, self.quad.permuteData))
        abssort = lambda x: x[abs(x).argsort()]   # sort: For the sake of the more precision
        for i in range(self.pp):
            ip = np.delete(self.ip, i)
            denominator = abssort(self.ip[i] - ip)            # interpolate points
            for j in range(self.quad.permuteData):
                numerator = abssort(self.quad.points[j] - ip)   # quadrature points
                mat[i][j] = (numerator/denominator).prod()
        return mat

    def mathbbH_1(self):
        mat = np.zeros((self.pp, self.quad.permuteData))
        # sort: For the sake of the more precision
        abssort = lambda x: x[abs(x).argsort()]   
        if self.pp == 2:
            mat[0] = -0.5
            mat[1] = 0.5
            return mat
        for i in range(self.pp):
            ip = np.delete(self.ip, i)
            denominator = abssort(self.ip[i] - ip)     # interpolate points
            for j in range(self.quad.permuteData):
                numerator = self.quad.points[j] - ip            # quadrature points
                tmp = 0.0
                for k in range(self.deg):
                    subnu = numerator.copy()              # for deepcopy
                    subnu[k] = 1.0
                    subnu = abssort(subnu)
                    tmp += (subnu/denominator).prod()
                mat[i][j] = tmp
        return mat



u = lambda x: np.sin(np.array(x))
f = lambda x: np.cos(np.array(x))
a = 0

## interpolation points

ip = Quadrature(k+1).getLobattoPiontsWeights()[0]


h = 1.0/nElement
eledof = k + 1
uh = np.zeros(eledof*nElement)
quad = Quadrature(k+6)
quad.getLobattoPiontsWeights()
itp = Interp(ip, quad)


phiphix = -1.0*np.tensordot(quad.weights*itp.matrixH_1, itp.matrixH.T, 1)
# phiphi = np.tensordot(quad.weights*itp.matrixH, itp.matrixH.T, 1)


vectorU = np.zeros(eledof)
vectorU[-1] = a
ja = 0.5*h

for ic in range(nElement):
    L, R = ic*h, (ic+1)*h
    x = 0.5*(L+R) + 0.5*quad.points*(R-L)
    tmp = (quad.weights)*f(x)
    vectorF = np.dot(itp.matrixH, tmp) * ja
    matrixA = phiphix.copy()
    matrixA[-1][-1] += 1
    vectorF[0] += vectorU[-1]
    vectorU = solve(matrixA, vectorF)
    uh[ic*eledof:(ic+1)*eledof] = vectorU


def error(uh, u):
    pts, wts = quad.points, quad.weights
    L1sum = L2sum = Lmax = 0.0
    for ic in range(nElement):
        L, R = ic*h, (ic+1)*h
        xele = 0.5*(L+R) + 0.5*(R-L)*pts
        uhele = uh[ic*eledof: (ic+1)*eledof]
        tmp = u(xele) - np.dot(uhele, itp.matrixH)
        tmp = abs(tmp)
        L1sum += (tmp*wts).sum()*h*0.5
        L2sum += (tmp*tmp*wts).sum()*h*0.5
        if(Lmax <= tmp.max()):
            Lmax = tmp.max()
    return L1sum, np.sqrt(L2sum), Lmax


ui = np.zeros(eledof*nElement)
for ic in range(nElement):
    L, R = ic*h, (ic+1)*h
    xele = 0.5*(L+R) + 0.5*(R-L)*ip
    ui[ic*eledof:(ic+1)*eledof] = u(xele)

print("h, k = ", h, k)
print("ui L1, L2, Loo error")
print(error(ui, u))
print("uh L1, L2, Loo error")
print(error(uh, u))


# import matplotlib.pyplot as plt
# fig = plt.figure(figsize=(8,8), dpi=72,facecolor="white")
# axes = plt.subplot(111)
# plt.axis([0, 1, -0.1, 1.1])
# for ic in range(nElement):
    # L, R = ic*h, (ic+1)*h
    # uhele = uh[ic*eledof: (ic+1)*eledof]
    # xele = 0.5*(L+R) + 0.5*(R-L)*quad.points
    # itpuhele = np.dot(uhele, itp.matrixH)
    # plt.plot(xele, itpuhele)
    # plt.plot(xele, u(xele), color='red')
# plt.show()